from os import sep
import typer
import pathlib
from kghub_downloader.download_utils import download_from_yaml
from monarch_gene_mapping.cli_utils import generate_gene_mappings
from monarch_gene_mapping.uniprot_idmapping_preprocess import filter_uniprot_id_mapping_file

typer_app = typer.Typer()


@typer_app.command(name="download")
def _download():
    download_from_yaml(
        yaml_file="monarch_gene_mapping/download.yaml",
        output_dir=".",
    )


@typer_app.command(name="preprocess-uniprot")
def preprocess_uniprot_mappings(
    directory: str = typer.Option(f"..{sep}data{sep}uniprot", help="Data Directory"),
    source_filename: str = typer.Option("idmapping_selected.tab", help="Target File Name"),
    target_filename: str = typer.Option("idmapping_filtered.tsv", help="Target File Name"),
    number_of_lines: int = typer.Option(0, help="Number of Lines"),
):
    filter_uniprot_id_mapping_file(
        directory=directory,
        source_filename=source_filename,
        target_filename=target_filename,
        number_of_lines=number_of_lines,
    )


@typer_app.command()
def generate(
    output_dir=typer.Option("output", help="Output directory"),
    download: bool = typer.Option(False, help="Pass to first download required data"),
    preprocess_uniprot: bool = typer.Option(False, help="Filter out UniProt ID mapping data after download"),
):
    if download:
        _download()
        print("\nData download complete!\n")

    # prefilter 'target' taxa in Uniprot data
    if preprocess_uniprot:
        filter_uniprot_id_mapping_file(
            directory="data/uniprot",
            source_filename="idmapping_selected.tab",
            target_filename="idmapping_filtered.tsv",
            number_of_lines=0,
        )

    print("\nGenerating gene mapping...\n")
    pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)
    mappings = generate_gene_mappings()

    print(f"\nResults saved in {output_dir}/gene_mappings.sssom.tsv")
    mappings.to_csv(f"{output_dir}/gene_mappings.sssom.tsv", sep="\t", index=False)


if __name__ == "__main__":
    typer_app()



from typing import List

import numpy as np
import pandas as pd
from pandas.core.frame import DataFrame

# The UniProtKB mapping tsv file lacks a header line
UNIPROT_ID_MAPPING_SELECTED_COLUMNS = [
    "UniProtKB-AC",
    "UniProtKB-ID",
    "GeneID",
    "RefSeq",
    "GI",
    "PDB",
    "GO",
    "UniRef100",
    "UniRef90",
    "UniRef50",
    "UniParc",
    "PIR",
    "NCBI-taxon",
    "MIM",
    "UniGene",
    "PubMed",
    "EMBL",
    "EMBL-CDS",
    "Ensembl",
    "Ensembl_TRS",
    "Ensembl_PRO",
    "Additional PubMed",
]


def add_prefix(prefix: str, column: pd.Series) -> pd.Series:
    """
    Add a prefix to all values in a series
    :param prefix: Prefix for values in series
    :param column: Series of values for prefixing
    :return:
    """
    return prefix + column.astype("str")


def df_mappings(
    df: DataFrame,
    subject_column: str,  # "GeneID"
    object_column: str,  # "Ensembl_gene_identifier"
    predicate_id: str = "skos:exactMatch",
    entity_delimiter: str = ";",
    mapping_justification: str = "semapv:UnspecifiedMatching",
    filter_column: str = None,  # "#tax_id",
    subject_curie_prefix: str = None,  # "NCBIGene:"
    object_curie_prefix: str = None,  # "ENSEMBL:"
    filter_ids: List[int] = None,  # [9031]
) -> DataFrame:
    """
    Create specified mappings from DataFrame
    :param df: DataFrame source for mapping values
    :param subject_column: Column containing subject ID's
    :param object_column: Column containing object ID's
    :param predicate_id: String for predicate ID
    :param entity_delimiter: Delimiter for splitting IDs
    :param mapping_justification: String for mapping justification
    :param filter_column: Column to filter DataFrame on
    :param filter_ids:
    :param subject_curie_prefix: Optional curie for prefixing subject ID's
    :param object_curie_prefix: Optional curie for prefixing object ID's
    :return:
    """
    # Filtering could be extracted and done before passing df but I think there is value keeping it here.
    if (filter_column is not None) and isinstance(filter_ids, list):
        # Create a copy to guarantee we aren't working on a slice
        df_filtered = df.loc[df[filter_column].isin(filter_ids), :].copy()
    else:
        # Create copy so we don't modify the original DataFrame
        df_filtered = df.copy()

    df_filtered["predicate_id"] = predicate_id
    df_filtered["mapping_justification"] = mapping_justification

    columns = {subject_column: "subject_id", object_column: "object_id"}
    select_columns = ["subject_id", "predicate_id", "object_id", "mapping_justification"]
    df_select = df_filtered.rename(columns=columns).loc[:, select_columns].copy()

    # Create a copy of the DataFrame with unmapped values
    # df_unmapped = df_select[df_select['subject_id'].isna() | df_select['object_id'].isna()]

    # Drop rows with missing values
    df_select = df_select.dropna(subset=["subject_id", "object_id"], how="any")
    df_select = df_select.drop_duplicates()

    # Expand rows with semicolon in subject_id or object_id to multiple rows
    df_select = explode_column(df_select, "subject_id", entity_delimiter)
    df_select = explode_column(df_select, "object_id", entity_delimiter)

    if subject_curie_prefix is not None:
        df_select["subject_id"] = add_prefix(subject_curie_prefix, df_select["subject_id"])
    if object_curie_prefix is not None:
        df_select["object_id"] = add_prefix(object_curie_prefix, df_select["object_id"])

    df_map = df_select.drop_duplicates().dropna().copy()
    return df_map  # , df_unmapped


def explode_column(df: DataFrame, column: str, delimiter: str) -> DataFrame:
    """
    Expand columns with delimiter separated lists to multiple rows for each
    :param df: DataFrame for column expansion
    :param column: Column name for expansion
    :param delimiter: Delimiter for splitting column
    :return:
    """
    # cast non-null items in column to string
    df[column] = np.where(pd.isnull(df[column]),df[column],df[column].astype(str))

    assign_kwargs = {column: df[column].str.split(delimiter)}
    df_exploded = df.assign(**assign_kwargs).explode(column).copy()

    # remove whitespace
    df_exploded[column] = df_exploded[column].str.strip()
    return df_exploded


def preprocess_alliance_df(
    df: DataFrame, exclude_taxon: List, include_curie: List, include_xref_curie: List
) -> DataFrame:
    taxon_filter = ~df["TaxonID"].isin(exclude_taxon)
    curie_filter = df["GeneID"].str.contains("|".join(include_curie))
    self_filter = df["GeneID"] != df["GlobalCrossReferenceID"]
    xref_curie_filter = df["GlobalCrossReferenceID"].str.startswith(tuple(include_xref_curie))
    df.loc[:, "GlobalCrossReferenceID"] = df["GlobalCrossReferenceID"].str.replace("NCBI_Gene:", "NCBIGene:")
    df_filtered = df.loc[taxon_filter & curie_filter & self_filter & xref_curie_filter, :]
    return df_filtered.copy()


def alliance_mapping() -> DataFrame:
    alliance_file = "data/alliance/GENECROSSREFERENCE_COMBINED.tsv.gz"
    alliance_df = pd.read_csv(alliance_file, sep="\t", dtype="string", comment="#")
    alliance_df_filtered = preprocess_alliance_df(
        df=alliance_df,
        exclude_taxon=["NCBITaxon:9606", "NCBITaxon:2697049"],
        include_curie=["MGI:", "RGD:", "FB:", "WB:", "ZFIN:", "Xenbase:"],
        include_xref_curie=["ENSEMBL:", "NCBI_Gene:", "UniProtKB:"],
    )
    alliance_mappings = df_mappings(
        df=alliance_df_filtered,
        subject_column="GeneID",
        subject_curie_prefix="",
        object_column="GlobalCrossReferenceID",
        object_curie_prefix="",
        predicate_id="skos:exactMatch",
        mapping_justification="semapv:UnspecifiedMatching",
    )

    return alliance_mappings


def generate_gene_mappings() -> DataFrame:
    mapping_dataframes = []

    ### Alliance mappings
    print("Generating Alliance mappings...")
    alliance_mappings = alliance_mapping()
    print(f"Generated {len(alliance_mappings)} Alliance mappings")
    assert len(alliance_mappings) > 400000
    mapping_dataframes.append(alliance_mappings)

    ### HGNC mappings

    print("\nGenerating HGNC to NCBI Gene mappings...")
    hgnc_df = pd.read_csv("data/hgnc/hgnc_complete_set.txt", sep="\t", dtype="string")
    hgnc_to_ncbi = df_mappings(
        df=hgnc_df,
        subject_column="hgnc_id",
        object_column="entrez_id",
        object_curie_prefix="NCBIGene:",
        predicate_id="skos:exactMatch",
        mapping_justification="semapv:UnspecifiedMatching",
    )
    print(f"Generated {len(hgnc_to_ncbi)} HGNC-NCBI Gene mappings")
    assert len(hgnc_to_ncbi) > 40000
    mapping_dataframes.append(hgnc_to_ncbi)

    print("\nGenerating HGNC to OMIM mappings...")
    hgnc_to_omim = df_mappings(
        df=explode_column(hgnc_df, "omim_id", "|"),
        subject_column="hgnc_id",
        object_column="omim_id",
        object_curie_prefix="OMIM:",
        predicate_id="skos:exactMatch",
        mapping_justification="semapv:UnspecifiedMatching",
    )
    print(f"Generated {len(hgnc_to_omim)} HGNC-OMIM mappings")
    assert len(hgnc_to_omim) > 16000
    mapping_dataframes.append(hgnc_to_omim)

    print("\nGenerating HGNC to UniProtKB mappings...")
    hgnc_to_uniprot = df_mappings(
        df=explode_column(hgnc_df, "uniprot_ids", "|"),
        subject_column="hgnc_id",
        object_column="uniprot_ids",
        object_curie_prefix="UniProtKB:",
        predicate_id="skos:closeMatch",
        mapping_justification="semapv:UnspecifiedMatching",
    )
    print(f"Generated {len(hgnc_to_uniprot)} HGNC-UniProtKB mappings")
    assert len(hgnc_to_uniprot) > 20000
    mapping_dataframes.append(hgnc_to_uniprot)

    print("\nGenerating HGNC to ENSEMBL Gene mappings...")
    hgnc_to_ensemble = df_mappings(
        df=hgnc_df,
        subject_column="hgnc_id",
        object_column="ensembl_gene_id",
        object_curie_prefix="ENSEMBL:",
        predicate_id="skos:exactMatch",
        mapping_justification="semapv:UnspecifiedMatching",
    )
    print(f"Generated {len(hgnc_to_ensemble)} HGNC-ENSEMBL Gene mappings")
    assert len(hgnc_to_ensemble) > 40000
    mapping_dataframes.append(hgnc_to_ensemble)

    ### NCBI mappings

    print("\nGenerating NCBIGene to ENSEMBL Gene mappings...")
    ensembl_df = pd.read_csv("data/ncbi/gene2ensembl.gz", compression="gzip", sep="\t")
    ensembl_to_ncbi = df_mappings(
        df=ensembl_df,
        subject_column="GeneID",
        subject_curie_prefix="NCBIGene:",
        object_column="Ensembl_gene_identifier",
        object_curie_prefix="ENSEMBL:",
        predicate_id="skos:exactMatch",
        mapping_justification="semapv:UnspecifiedMatching",
        filter_column="#tax_id",
        # Chicken: 9031, Dog: 9615, Cow, 9913, Pig: 9823, Aspergillus ('Emericella') nidulans: 227321
        filter_ids=[9031, 9615, 9913, 9823, 227321],
    )
    print(f"Generated {len(ensembl_to_ncbi)} ENSEMBL-NCBIGene Gene mappings")
    assert len(ensembl_to_ncbi) > 70000
    mapping_dataframes.append(ensembl_to_ncbi)

    ### UniProtKB mappings

    print("\nGenerating UniProtKB to NCBI Gene mappings...")
    uniprot_df = pd.read_csv(
        "data/uniprot/idmapping_filtered.tsv.gz",  # filtered down to target species
        names=UNIPROT_ID_MAPPING_SELECTED_COLUMNS,
        compression="gzip",
        sep="\t",
        low_memory=False,
    )
    uniprot_to_ncbi = df_mappings(
        df=uniprot_df,
        subject_column="GeneID",
        subject_curie_prefix="NCBIGene:",
        predicate_id="skos:exactMatch",
        object_column="UniProtKB-AC",
        object_curie_prefix="UniProtKB:",
        mapping_justification="semapv:UnspecifiedMatching",
        filter_column="NCBI-taxon",
        # Chicken: 9031, Dog: 9615, Cow, 9913, Pig: 9823, Aspergillus ('Emericella') nidulans: 227321
        filter_ids=[9031, 9615, 9913, 9823, 227321],
        entity_delimiter=";",
    )
    print(f"Generated {len(uniprot_to_ncbi)} UniProtKB-NCBIGene Gene mappings")
    assert len(uniprot_to_ncbi) > 70000
    mapping_dataframes.append(uniprot_to_ncbi)

    mappings = pd.concat(mapping_dataframes)
    for row in mappings.itertuples():
        assert not ("<NA>" in row.subject_id)
        assert not ("<NA>" in row.object_id)
    return mappings



#!/usr/bin/env python
"""
Utility methods for pre-filtering of **huge** UniProdKB ID
mapping input file (i.e. data/uniprot/idmapping_selected.tab.gz)
The script takes about 9 minutes to filter the input dataset containing
~60 million original entries, down to ~1.2 million (of taxon-specific) entries.
"""
from os import sep

from sys import stderr
import argparse

from gzip import open as gzip_open, BadGzipFile

from datetime import datetime

from typing import Optional, List, Set

# Hard-coded list of target species (TODO: externalize this list in a configuration file?)
target_species: Set = {
    "9606",  # Homo sapiens
    "10090",  # Mus musculus (mouse)
    "9615",  # Canis lupus familiaris - domestic dog
    # "9685", # Felis catus - domestic cat
    "9913",  # Bos taurus - cow
    "9823",  # Sus scrofa - pig
    "10116",  # Rattus norvegicus (Norway rat)
    "9031",  # Gallus gallus
    "8364",  # Xenopus tropicalis - tropical clawed frog
    "7955",  # Danio rerio (Zebrafish)
    "7227",  # Drosophila melanogaster (fruit fly)
    "6239",  # Caenorhabditus elegans
    "44689",  # Dictylostelium
    "227321",  # Emericella nidulans (strain FGSC A4 etc.) (Aspergillus nidulans)
    "4896",  # Schizosaccharomyces pombe ("fission" yeast)
    "4932",  # Saccharomyces cerevisiae (baker's "budding" yeast)
}

# idmapping_selected.tab field 13 is column 12 in TSV array...
UNIPROT_ID_MAPPING_NCBI_TAXON_COLUMN = 12


def target_taxon(line: Optional[str]) -> bool:
    if not line:
        return False
    part: List[str] = line.split("\t")
    if part[UNIPROT_ID_MAPPING_NCBI_TAXON_COLUMN] in target_species:
        return True
    else:
        return False


PROGRESS_MONITOR_INCR: int = 1000


def filter_uniprot_id_mapping_file(
    directory: str, source_filename: str, target_filename: str, number_of_lines: int = 0  # root file name only?
) -> bool:
    """
    Filters contents of a UniProKB idmapping selected tab gzip'd archive against the target list of taxa.
    :param directory: str, location of source data file
    :param source_filename: str, root file name of input source data archive
    :param target_filename: str, root file name of output target data archive
    :param number_of_lines: int, positive number of lines parsed; 'all' lines parsed if omitted or set to zero
    :return: bool, True if filtering was successful; False if unsuccessful
    """
    if not directory:
        directory = "."
    directory_path: str = directory

    assert source_filename
    assert target_filename
    assert number_of_lines >= 0

    print(
        f"\nBegin file filtering '{number_of_lines if number_of_lines else 'all'}'"
        + f" lines in '{source_filename}' at {datetime.now().isoformat()}. "
        + f"\nPatience! This may take a little awhile!...\n"
    )

    # A standard gzip compressed TSV file
    source_gz_filename: str = f"{source_filename}.gz"
    source_gz_file_path: str = f"{directory_path}{sep}{source_gz_filename}"
    target_gz_filename: str = f"{target_filename}.gz"
    target_gz_file_path: str = f"{directory_path}{sep}{target_gz_filename}"
    n: int = 0
    p: int = 0  # visible progress monitor
    print("Processing... ")
    try:
        with gzip_open(source_gz_file_path, mode="rt", encoding="utf-8") as source_gz_file:
            with gzip_open(target_gz_file_path, mode="wt", encoding="utf-8") as target_file:
                for line in source_gz_file:
                    p += 1
                    if target_taxon(line):
                        print(line, file=target_file, end="")
                        n += 1
                    if number_of_lines and n > number_of_lines:
                        break
                    if not (p % PROGRESS_MONITOR_INCR):
                        print(f"{p} lines", end="\r")

    except FileNotFoundError as fnf:
        print(f"\nFile '{source_gz_file_path}' not found, at {datetime.now().isoformat()}", file=stderr)
        return False
    except BadGzipFile as bzf:
        print(
            f"\nFailed filtering '{source_filename}' at {datetime.now().isoformat()}: "
            + f"Gzip file access exception: {str(bzf)}",
            file=stderr,
        )
        return False

    print(f"\nFinished filtering file '{source_gz_file_path}' at {datetime.now().isoformat()}")
    return True


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--directory",
        default=f"..{sep}data{sep}uniprot",
        help="Working directory containing the data (default: 'data/uniprot')",
    )
    parser.add_argument(
        "-s",
        "--source",
        default="idmapping_selected.tab",
        help="Source root data file name (default: 'idmapping_selected.tab')",
    )
    parser.add_argument(
        "-t", "--target", default="idmapping_filtered.tsv", help="Target root data file name (default: 'idmapping_filtered.tsv')"
    )
    parser.add_argument(
        "-n",
        "--number_of_lines",
        type=int,
        default=0,
        help="Maximum (positive) number of lines to process; n == 0 implies 'process all' (default: 0 == 'all')",
    )

    args = parser.parse_args()

    if filter_uniprot_id_mapping_file(
        directory=args.directory,
        source_filename=args.source,
        target_filename=args.target,
        number_of_lines=args.number_of_lines,
    ):
        quantity: str = args.number_of_lines > 0 if args.number_of_lines else "All"
        print(
            f"\n{quantity} lines of file '{args.source}' in directory '{args.directory}'"
            + f" successfully filtered into file '{args.target}'"
        )
    else:
        # fail filtering
        print(f"\nERROR: file '{str(args.source)}' in directory '{args.directory}' could NOT be processed?")
